'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Lexer = undefined;

var _symbol = require('./symbol.js');

const $private = Symbol('private');
const noValue = null;

class Rule {
  constructor(fn) {
    this[$private] = { fn: fn };
  }

  test(str) {
    const res = this[$private].res = this.compare(str);

    if (!res || res.index !== 0) {
      this[$private].res = null;
      return null;
    }

    return res[0].length;
  }

  execute(lex, token) {
    return this[$private].fn.call(lex, this[$private].res, token);
  }
}

class StringRule extends Rule {
  constructor(str, fn) {
    super(fn);
    this[$private].str = str;
  }

  compare(str) {
    const res = [this[$private].str];
    res.index = str.indexOf(this[$private].str);

    return res.index < 0 ? null : res;
  }

  toString() {
    return 'string("' + this[$private].str + '")';
  }
}

class RegexpRule extends Rule {
  constructor(regexp, fn) {
    super(fn);
    this[$private].regexp = regexp;
  }

  compare(str) {
    return str.match(this[$private].regexp);
  }

  toString() {
    return 'regexp(' + this[$private].regexp + ')';
  }
}

class Lexer {
  constructor(input) {
    this[$private] = {
      rules: { default: [] }
    };
    this.reset(input);
  }

  reset(input) {
    this[$private].token = { sy: new _symbol.LngSymbol() };
    this[$private].input = input;
    this[$private].currentLine = this[$private].input.substr(0, this[$private].input.indexOf('\n'));
    this[$private].pos = { line: 1, col: 1 };
    this[$private].scopes = [this[$private].rules.default];
  }

  getLookAhead() {
    if (this[$private].token.sy.type === _symbol.LngSymbol.epsilon) this.eat();

    return this[$private].token;
  }

  eat() {
    this.$eat(true);
  }

  registerRule(str, fn, ...scopes) {
    scopes.forEach(scope => {
      if (!(scope in this[$private].rules)) this[$private].rules[scope] = [];
      if (str instanceof RegExp) this[$private].rules[scope].push(new RegexpRule(str, fn));else this[$private].rules[scope].push(new StringRule(str, fn));
    });
  }

  pushScope(scopeName) {
    if (!(scopeName in this[$private].rules)) throw new Error('Trying to push an invalid scope `' + scopeName + "'");

    this[$private].scopes.push(this[$private].rules[scopeName]);
  }

  popScope() {
    if (this[$private].scopes.length <= 1) throw new Error('Trying to pop a scope but there is only default in the stack.');
    this[$private].scopes.pop();
  }

  $eat(reset) {
    if (reset) this[$private].token = { val: noValue, sy: new _symbol.LngSymbol(_symbol.LngSymbol.epsilon) };

    const scope = this[$private].scopes[this[$private].scopes.length - 1];

    let bestSize = 0;
    let bestRule = null;
    scope.forEach(rule => {
      let size = rule.test(this[$private].input);
      if (size <= bestSize && (size === null || bestRule)) return;
      bestSize = size;
      bestRule = rule;
    });

    if (!this[$private].input.length && bestRule === null) {
      this[$private].token.sy = new _symbol.LngSymbol(_symbol.LngSymbol.token, 'EOF');
      return;
    }

    if (bestRule === null) {
      console.error("Character '" + this[$private].input[0] + "' jumped because no rule matched current input.");
      this.move(1);
      return this.eat();
    }

    this.hasMoved = false;
    this[$private].token.location = { line: this[$private].pos.line, col: this[$private].pos.col, length: bestSize };

    const res = bestRule.execute(this, this[$private].token);

    if (!this.hasMoved) this.move(bestSize);

    if (!res) this.$eat(false);
  }

  move(size) {
    this.hasMoved = true;
    const str = this[$private].input.substr(0, size);
    this[$private].input = this[$private].input.substr(size);

    const lines = str.split('\n');
    if (lines.length < 2) {
      this[$private].pos.col += str.length;
      return this;
    }

    this[$private].pos.col = lines[lines.length - 1].length + 1;
    this[$private].pos.line += lines.length - 1;
    return this;
  }

  get line() {
    return this[$private].pos.line;
  }

  get col() {
    return this[$private].pos.col;
  }
}

exports.Lexer = Lexer; /*
                       const descr = Object.getOwnPropertyDescriptor(Lexer.prototype, '$eat')
                       descr.enumerable = false
                       Object.defineProperty(Lexer.prototype, '$eat', descr)
                       //*/