'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.PrettyPrinter = undefined;

var _chalk = require('chalk');

var _chalk2 = _interopRequireDefault(_chalk);

var _ast = require('../ast');

var ast = _interopRequireWildcard(_ast);

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class Printer {
  constructor() {
    this.str = '';
    this.indentChars = '';
    this.cols = [];
  }

  indent(indent = '  ') {
    this.indentChars += indent;
    return this;
  }

  deindent(indent = '  ') {
    this.indentChars = this.indentChars.substr(0, this.indentChars.length - indent.length);
    return this;
  }

  print(value) {
    const dest = this.cols.length ? this.row : this;

    if (dest.str[dest.str.length - 1] === '\n') dest.str += this.indentChars;
    dest.str += value;
    return this;
  }

  linefeed() {
    const dest = this.cols.length ? this.row : this;
    dest.str += '\n';
    return this;
  }

  get col() {
    return this.cols[this.cols.length - 1];
  }

  get row() {
    return this.col[this.col.length - 1];
  }

  startcol() {
    this.cols.push([]);
  }

  newrow(header) {
    this.col.push({ header: header, headerLength: _chalk2.default.stripColor(header).length, str: '' });
  }

  endcol() {
    const rows = this.cols.pop();
    const pad = rows.reduce((acc, row) => acc > row.headerLength ? acc : row.headerLength);
    rows.forEach(row => {
      for (var i = 0; i < pad - row.headerLength; ++i) this.print(' ');

      let indent = ' ';
      for (var i = 0; i < pad; ++i) indent += ' ';

      this.print(row.header + ' ');

      this.print(row.str.replace(/\n(.)/g, '\n' + indent + '$1'));
    });
  }

  value() {
    return this.str;
  }
}

class PrettyPrinter {
  instructionArray(arr, addLinefeed = true) {
    arr.forEach(instruction => {
      instruction.accept(this);
      if (addLinefeed) this.str.linefeed();
    });
  }

  visitAlgorithm(algo) {
    this.str = new Printer();

    this.str.print(_chalk2.default.yellow('algorithme ') + _chalk2.default.white(algo.name)).linefeed().indent();
    algo.decls.accept(this);
    this.str.deindent().print(_chalk2.default.yellow('debut')).linefeed().indent();
    this.instructionArray(algo.instructions);
    this.str.deindent().print(_chalk2.default.yellow('fin algorithme ') + _chalk2.default.white(algo.name));

    return this.str.value();
  }

  visitAlgoDecls(decls) {
    if (decls.consts.length) {
      this.str.print(_chalk2.default.yellow('constantes')).linefeed().indent();
      this.instructionArray(decls.consts);
      this.str.deindent().linefeed();
    }
    if (decls.types.length) {
      this.str.print(_chalk2.default.yellow('types')).linefeed().indent();
      this.instructionArray(decls.types);
      this.str.deindent().linefeed();
    }
    if (decls.variables.length) {
      this.str.print(_chalk2.default.yellow('variables')).linefeed().indent();
      this.instructionArray(decls.variables);
      this.str.deindent().linefeed();
    }
    if (decls.routines && decls.routines.length) this.instructionArray(decls.routines);
  }

  visitConst(c) {
    this.str.print(_chalk2.default.white(c.name) + ' = ');
    c.value.accept(this);
  }

  visitEnum(e) {
    this.str.print(_chalk2.default.white(e.name) + _chalk2.default.magenta(' = ') + '(' + e.values.join(', ') + ')');
  }

  visitArray(array) {
    this.str.print(_chalk2.default.white(array.name) + _chalk2.default.magenta(' = ') + _chalk2.default.blue(array.dimensions.join(_chalk2.default.magenta(' x '))) + ' ' + _chalk2.default.white(array.type));
  }

  visitRecord(record) {
    this.str.print(_chalk2.default.white(record.name) + _chalk2.default.magenta(' = ') + _chalk2.default.yellow('enregistrement')).linefeed().indent();
    this.instructionArray(record.properties);
    this.str.deindent().print(_chalk2.default.yellow('fin enregistrement ') + _chalk2.default.white(record.name));
  }

  visitVariables(variables) {
    this.str.print(_chalk2.default.white(variables.type) + ' ' + variables.names.join(', '));
  }

  visitPointer(pointer) {
    this.str.print(_chalk2.default.white(pointer.name) + _chalk2.default.magenta(' = ↑') + _chalk2.default.white(pointer.pointOn));
  }

  visitFunction(func) {
    this.str.print(_chalk2.default.yellow('algorithme fonction ') + _chalk2.default.white(func.name) + _chalk2.default.magenta(': ') + _chalk2.default.white(func.type)).linefeed().indent();
    func.decls.accept(this);
    this.str.deindent().print(_chalk2.default.yellow('debut')).linefeed().indent();
    this.instructionArray(func.instructions);
    this.str.deindent().print(_chalk2.default.yellow('fin algorithme fonction ') + _chalk2.default.white(func.name)).linefeed();
  }

  visitProcedure(func) {
    this.str.print(_chalk2.default.yellow('algorithme procedure ') + _chalk2.default.white(func.name)).linefeed().indent();
    func.decls.accept(this);
    this.str.deindent().print(_chalk2.default.yellow('debut')).linefeed().indent();
    this.instructionArray(func.instructions);
    this.str.deindent().print(_chalk2.default.yellow('fin algorithme procedure ') + _chalk2.default.white(func.name)).linefeed();
  }

  visitRoutine(routine) {
    if (routine.params) routine.params.accept(this);
    this.visitAlgoDecls(routine);
  }

  visitParameters(parameters) {
    if (!parameters.reverseOrder && parameters.locals.length) {
      this.str.print(_chalk2.default.yellow('parametres locaux')).linefeed().indent();
      this.instructionArray(parameters.locals);
      this.str.deindent().linefeed();
    }

    if (parameters.globals.length) {
      this.str.print(_chalk2.default.yellow('parametres globaux')).linefeed().indent();
      this.instructionArray(parameters.globals);
      this.str.deindent().linefeed();
    }

    if (parameters.reverseOrder && parameters.locals.length) {
      this.str.print(_chalk2.default.yellow('parametres locaux')).linefeed().indent();
      this.instructionArray(parameters.locals);
      this.str.deindent().linefeed();
    }
  }

  visitBinary(binary) {
    binary.left.accept(this);
    this.str.print(' ' + _chalk2.default.magenta(binary.op) + ' ');
    binary.right.accept(this);
  }

  visitRecordAccess(record) {
    record.record.accept(this);
    this.str.print(_chalk2.default.magenta('.') + _chalk2.default.white(record.property));
  }

  visitIdentifierRef(identifier) {
    this.str.print(_chalk2.default.white(identifier.ref));
  }

  visitArrayAccess(array) {
    array.array.accept(this);
    this.str.print(_chalk2.default.magenta('['));
    array.positions.forEach((position, i) => {
      if (i != 0) this.str.print(', ');
      position.accept(this);
    });
    this.str.print(_chalk2.default.magenta(']'));
  }

  visitUnary(unary) {
    this.str.print(_chalk2.default.magenta(unary.op));
    if (unary.op === 'non') this.str.print(' ');
    unary.expr.accept(this);
  }

  printValue(value) {
    switch (value.type) {
      case 'chaine':
      case 'string':
        return _chalk2.default.green('"' + value.value.replace(/"/g, '\\"').replace(/\n/g, '\\n') + '"');
      case 'caractere':
        return _chalk2.default.green("'" + value.value.replace(/"/g, '\\"').replace(/\n/g, '\\n') + "'");
      default:
        return _chalk2.default.blue(value.value);
    }
  }

  visitValue(value) {
    this.str.print(this.printValue(value));
  }

  visitSi(si) {
    this.str.print(_chalk2.default.yellow('si '));
    si.expr.accept(this);
    this.str.print(_chalk2.default.yellow(' alors')).linefeed().indent();
    this.instructionArray(si.vrai);
    if (si.faux) {
      this.str.deindent().print(_chalk2.default.yellow('sinon')).linefeed().indent();
      this.instructionArray(si.faux);
    }
    this.str.deindent().print(_chalk2.default.yellow('fin si'));
  }

  visitRoutineCall(routine) {
    this.str.print(_chalk2.default.white(routine.ref) + '(');
    routine.args.forEach((arg, i) => {
      if (i != 0) this.str.print(', ');
      arg.accept(this);
    });
    this.str.print(')');
  }

  visitPour(pour) {
    this.str.print(_chalk2.default.yellow('pour ') + _chalk2.default.white(pour.ref) + _chalk2.default.magenta(' ← '));
    pour.start.accept(this);
    this.str.print(_chalk2.default.yellow(' jusqu\'a '));
    pour.end.accept(this);
    if (pour.desc) this.str.print(_chalk2.default.yellow(' decroissant'));
    this.str.print(_chalk2.default.yellow(' faire')).linefeed().indent();
    this.instructionArray(pour.loop);
    this.str.deindent().print(_chalk2.default.yellow('fin pour'));
  }

  visitTantQue(tantque) {
    this.str.print(_chalk2.default.yellow(tantque.oneLoop ? 'faire' : 'tant que '));
    if (!tantque.oneLoop) {
      tantque.expr.accept(this);
      this.str.print(_chalk2.default.yellow(' faire'));
    }
    this.str.linefeed().indent();
    this.instructionArray(tantque.loop);
    this.str.deindent().print(_chalk2.default.yellow(tantque.oneLoop ? 'tant que ' : 'fin tant que'));
    if (tantque.oneLoop) tantque.expr.accept(this);
  }

  visitPointerAccess(pointer) {
    pointer.pointer.accept(this);
    this.str.print(_chalk2.default.magenta('↑'));
  }

  visitSelon(selon) {
    this.str.print(_chalk2.default.yellow('selon '));
    selon.test.accept(this);
    this.str.print(_chalk2.default.yellow(' faire')).linefeed().indent();

    this.str.startcol();
    this.instructionArray(selon.expressions, false);

    if (selon.either && selon.either.length) {
      this.str.newrow(_chalk2.default.yellow('autrement'));
      this.instructionArray(selon.either);
    }

    this.str.endcol();

    this.str.deindent().print(_chalk2.default.yellow('fin selon'));
  }

  visitSelonExpression(expr) {
    const header = expr.scalars.map(scalar => scalar.value instanceof ast.IdentifierRef ? scalar.value.ref : this.printValue(scalar.value)).join(', ');
    this.str.newrow(header + _chalk2.default.yellow(':'));
    this.instructionArray(expr.expressions);
  }
}
exports.PrettyPrinter = PrettyPrinter;