import {Ast} from './base.js'
import {Routine} from './routine.js'
import {JungleError} from '../jungle/error.js'

export class Algorithm extends Ast {
  constructor(location, name, decls, instructions) {
    super(location)
    this.name = name
    this.decls = decls
    this.instructions = instructions
  }

  getParamDeclsInOrder() {
    const params = []

    if (!this.decls.params.reverseOrder && this.decls.params.locals.length)
      params.push(...this.decls.params.locals)
    if (this.decls.params.globals.length)
      params.push(...this.decls.params.globals)
    if (this.decls.params.reverseOrder && this.decls.params.locals.length)
      params.push(...this.decls.params.locals)

    return params
  }

  getOverVariables(variables, callback) {
    let index = 0
    variables.forEach(variables => {
      variables.names.forEach((name, subIndex) => callback(variables, name, index + subIndex))
      index += variables.names.length
    })
  }

  checkParameters(args) {
    if (!(this.decls instanceof Routine))
      throw new JungleError('Cannot check parameters on a non routine algorithm…')

    const paramsDecls = this.getParamDeclsInOrder()

    const nbArgs = paramsDecls.reduce((acc, v) => acc + v.names.length, 0)
    if (nbArgs !== args.length)
      throw new JungleError('la routine ' + this.name + ' prend ' + nbArgs + ' paramètres mais ' + args.length +
        ' paramètres ont été fournis')

    this.getOverVariables(paramsDecls, (variables, name, index) => {
      if (!variables.type.isCompatibleWith(args[index].resType))
        throw new JungleError('L\'argument ' + (index + 1) + ' de la routine routine ' + this.name + ' attend un ' +
          variables.type.name + ' mais un ' + args[index].resType.name + ' a été fourni à la place', args[index].location)

      if (variables.assignable && !args[index].writable)
        throw new JungleError('L\'argument ' + (index + 1) + ' de la routine ' + this.name + ' attend une expression ' +
          'assignable pour un paramètre global', args[index].location)
    })
  }

  execute(executor, params) {
    if (this.decls instanceof Routine) {
      this.getOverVariables(
        this.getParamDeclsInOrder(),
        (variables, name, index) => {
          if (variables.assignable)
            return executor.variables.alias(name, params[index])
          executor.variables.add(name, variables.type).value = variables.type.create(params[index].value)
        }
      )
    }

    executor.array(this.decls.variables)
    executor.array(this.instructions)

    if (this.type && this.type.name !== '<no value>' && !this.returnValue)
      throw new JungleError('retourne n\'a pas été appelé durant l\'exécution de la routine ' + this.name)

    const res = this.returnValue
    this.returnValue = null
    return res
  }

  createReturn(value) {
    return {
      type: this.type,
      get value() {
        return value
      },
    }
  }

  stop(value) {
    this.returnValue = this.createReturn(value)
  }
}
