import {LngSymbol} from './symbol.js'
import {Lexer} from './lexer.js'
import {JungleError} from './error.js'

const $private = Symbol('private')

class ParsingTables {
  constructor(states) {
    this[$private] = {states: states}
  }

  getAction(i, token) {
    if (i >= this[$private].states.actions.length)
      throw new JungleError('State ' + state + ' is unkown.', token.location)
    return this[$private].states.actions[i].getRule(token)
  }

  getGoto(i, token) {
    if (i >= this[$private].states.gotos.length)
      throw new JungleError('State ' + state + ' is unkown.', token.location)
    return this[$private].states.gotos[i].getRule(token)
  }
}

class State {
  constructor(rules) {
    this[$private] = {rules: rules}
  }

  getRule(token) {
    var res = this[$private].rules[token.sy.name]

    if (!res)
      throw new JungleError('Symbol (' + LngSymbol.formatedName(token.sy.name) + ') is invalid. Possibles symbols are: ' +
        Object.keys(this[$private].rules).map(LngSymbol.formatedName).join(', '), token.location)

    return res.action
  }
}

class Action {
  constructor(symbol, action) {
    this[$private] = {symbol: symbol, action: action}
  }

  get symbol() {
    return this[$private].symbol
  }

  get type() {
    return this[$private].action.type
  }

  get action() {
    return this[$private].action
  }
}

class Result {
  constructor(val) {
    this[$private] = {val: val}
  }

  get val() {
    return this[$private].val
  }
}

function createAction([symbol, actionDescr]) {
  switch(actionDescr[0]) {
  case 's':
    return new Action(symbol, {
      type: 'shift',
      to: +actionDescr.substr(1),
    })
  case 'r':
    const matches = actionDescr.match(/^r\((.*?)\[(.*?)\],(.*?)\)$/)
    return new Action(symbol, {
      type: 'reduce',
      rule: matches[1],
      choice: matches[2],
      pop: matches[3],
    })
  case 'a':
    return new Action(symbol, {
      type: 'accept'
    })
  case 'g':
    return new Action(symbol, {
      type: 'goto',
      to: +actionDescr.substr(1),
    })
  default:
    throw new Error('Unknown rule: ' + actionDescr)
  }
}

export class Parser {
  constructor(ptStr, BaseLexer) {
    const [actions, gotos] = ptStr.split('<||||>')
      .map(states => states.split('<|||>')
        .map(items => items &&
          items.split('<||>')
            .reduce((acc, item) => {
              const action = createAction(item.split('<|>'))
              acc[action.symbol] = action
              return acc
            }, {}) || {}
        )
        .map(state => new State(state))
      )

    this[$private] = {
      tables: new ParsingTables({actions: actions, gotos: gotos}),
      fns: {},
      tokenStack: [],
      stateStack: [],
      BaseLexer: BaseLexer,
    }
  }

  acceptAction() {
    const tokenStack = this[$private].tokenStack

    throw new Result(tokenStack[tokenStack.length - 1].val)
  }

  shiftAction(lexer, action) {
    this[$private].stateStack.push(action.to)
    this[$private].tokenStack.push(lexer.getLookAhead())
    lexer.eat()
  }

  reduceAction(lexer, action) {
    const stateStack = this[$private].stateStack
    const tokenStack = this[$private].tokenStack

    let vals = []
    for (let i = 0; i < action.pop; ++i) {
      const token = tokenStack[tokenStack.length - 1]
      vals.unshift({location: token.location, value: token.val})
      tokenStack.pop()
      stateStack.pop()
    }

    var token = {
      sy: new LngSymbol(LngSymbol.rule, action.rule),
      val: this.reduce(action.rule, action.choice, vals),
    }

    tokenStack.push(token)

    this.visit(lexer, this[$private].tables
      .getGoto(stateStack[stateStack.length - 1], tokenStack[tokenStack.length - 1])
    )
  }

  gotoAction(lexer, action) {
    const stateStack = this[$private].stateStack

    stateStack.push(action.to)
  }

  visit(lexer, action) {
    this[action.type + 'Action'](lexer, action)
  }

  parse(strOrLexer) {
    if (!(strOrLexer instanceof Lexer))
      strOrLexer = new this[$private].BaseLexer(strOrLexer);

    try {
      this.$parse(strOrLexer, false)
    }
    catch (err) {
      if (!(err instanceof Result))
        throw err

      return err.val
    }
  }

  check(strOrLexer) {
    if (!(strOrLexer instanceof Lexer))
      strOrLexer = new this[$private].BaseLexer(strOrLexer);

    try {
      this.$parse(strOrLexer, true)
    }
    catch (err) {
      return err instanceof Result
    }

    return false
  }

  registerCode(rule, choice, fn) {
    if (!(rule in this[$private].fns))
      this[$private].fns[rule] = []
    this[$private].fns[rule][choice] = fn
  }

  $parse(lexer, checkOnly) {
    const stateStack = this[$private].stateStack
    const tokenStack = this[$private].tokenStack

    const stateLength = stateStack.length
    const tokenLength = tokenStack.length

    stateStack.push(0)

    try {
      while (true) {
        this.visit(lexer, this[$private].tables
          .getAction(stateStack[stateStack.length - 1], lexer.getLookAhead())
        )
      }
    }
    catch (err) {
      stateStack.splice(stateLength, stateStack.length)
      tokenStack.splice(tokenLength, tokenStack.length)
      throw err
    }
  }

  reduce(rule, choice, values) {
    if (this[$private].checkOnly)
      return null

    if (!(rule in this[$private].fns))
      throw new Error('No reduction function found for rule "' + rule + '"')
    if (this[$private].fns[rule].length <= choice)
      throw new Error('No reduction function found for rule "' + rule + '[' + choice + ']"')

    return this[$private].fns[rule][choice].apply(this, values)
  }
}
