import chalk from 'chalk'
import * as ast from '../ast'

class Printer {
  constructor() {
    this.str = ''
    this.indentChars = ''
    this.cols = []
  }

  indent(indent = '  ') {
    this.indentChars += indent
    return this
  }

  deindent(indent = '  ') {
    this.indentChars = this.indentChars.substr(0, this.indentChars.length - indent.length)
    return this
  }

  print(value) {
    const dest = this.cols.length ? this.row : this;

    if (dest.str[dest.str.length - 1] === '\n')
      dest.str += this.indentChars
    dest.str += value
    return this
  }

  linefeed() {
    const dest = this.cols.length ? this.row : this;
    dest.str += '\n'
    return this
  }

  get col() {
    return this.cols[this.cols.length - 1]
  }

  get row() {
    return this.col[this.col.length - 1]
  }

  startcol() {
    this.cols.push([])
  }

  newrow(header) {
    this.col.push({header: header, headerLength: chalk.stripColor(header).length, str: ''})
  }

  endcol() {
    const rows = this.cols.pop()
    const pad = rows.reduce((acc, row) => acc > row.headerLength ? acc : row.headerLength)
    rows.forEach(row => {
      for (var i = 0; i < pad - row.headerLength; ++i)
        this.print(' ')

      let indent = ' '
      for (var i = 0; i < pad; ++i)
        indent += ' '

      this.print(row.header + ' ')

      this.print(row.str.replace(/\n(.)/g, '\n' + indent + '$1'))
    })
  }

  value() {
    return this.str
  }
}

export class PrettyPrinter {
  instructionArray(arr, addLinefeed = true) {
    arr.forEach(instruction => {
      instruction.accept(this)
      if (addLinefeed)
        this.str.linefeed()
    })
  }

  visitAlgorithm(algo) {
    this.str = new Printer()

    this.str.print(chalk.yellow('algorithme ') + chalk.white(algo.name)).linefeed().indent()
    algo.decls.accept(this)
    this.str.deindent().print(chalk.yellow('debut')).linefeed().indent()
    this.instructionArray(algo.instructions)
    this.str.deindent().print(chalk.yellow('fin algorithme ') + chalk.white(algo.name))

    return this.str.value()
  }

  visitAlgoDecls(decls) {
    if (decls.consts.length) {
      this.str.print(chalk.yellow('constantes')).linefeed().indent()
      this.instructionArray(decls.consts)
      this.str.deindent().linefeed()
    }
    if (decls.types.length) {
      this.str.print(chalk.yellow('types')).linefeed().indent()
      this.instructionArray(decls.types)
      this.str.deindent().linefeed()
    }
    if (decls.variables.length) {
      this.str.print(chalk.yellow('variables')).linefeed().indent()
      this.instructionArray(decls.variables)
      this.str.deindent().linefeed()
    }
    if (decls.routines && decls.routines.length)
      this.instructionArray(decls.routines)
  }

  visitConst(c) {
    this.str.print(chalk.white(c.name) + ' = ')
    c.value.accept(this)
  }

  visitEnum(e) {
    this.str.print(chalk.white(e.name) + chalk.magenta(' = ') + '(' + e.values.join(', ') + ')')
  }

  visitArray(array) {
    this.str.print(chalk.white(array.name) + chalk.magenta(' = ') + chalk.blue(array.dimensions.join(chalk.magenta(' x '))) + ' ' + chalk.white(array.type))
  }

  visitRecord(record) {
    this.str.print(chalk.white(record.name) + chalk.magenta(' = ') + chalk.yellow('enregistrement')).linefeed().indent()
    this.instructionArray(record.properties)
    this.str.deindent().print(chalk.yellow('fin enregistrement ') + chalk.white(record.name))
  }

  visitVariables(variables) {
    this.str.print(chalk.white(variables.type) + ' ' + variables.names.join(', '))
  }

  visitPointer(pointer) {
    this.str.print(chalk.white(pointer.name) + chalk.magenta(' = ↑') + chalk.white(pointer.pointOn))
  }

  visitFunction(func) {
    this.str.print(chalk.yellow('algorithme fonction ') + chalk.white(func.name) + chalk.magenta(': ') + chalk.white(func.type)).linefeed().indent()
    func.decls.accept(this)
    this.str.deindent().print(chalk.yellow('debut')).linefeed().indent()
    this.instructionArray(func.instructions)
    this.str.deindent().print(chalk.yellow('fin algorithme fonction ') + chalk.white(func.name)).linefeed()
  }

  visitProcedure(func) {
    this.str.print(chalk.yellow('algorithme procedure ') + chalk.white(func.name)).linefeed().indent()
    func.decls.accept(this)
    this.str.deindent().print(chalk.yellow('debut')).linefeed().indent()
    this.instructionArray(func.instructions)
    this.str.deindent().print(chalk.yellow('fin algorithme procedure ') + chalk.white(func.name)).linefeed()
  }

  visitRoutine(routine) {
    if (routine.params)
      routine.params.accept(this)
    this.visitAlgoDecls(routine)
  }

  visitParameters(parameters) {
    if (!parameters.reverseOrder && parameters.locals.length) {
      this.str.print(chalk.yellow('parametres locaux')).linefeed().indent()
      this.instructionArray(parameters.locals)
      this.str.deindent().linefeed()
    }

    if (parameters.globals.length) {
      this.str.print(chalk.yellow('parametres globaux')).linefeed().indent()
      this.instructionArray(parameters.globals)
      this.str.deindent().linefeed()
    }

    if (parameters.reverseOrder && parameters.locals.length) {
      this.str.print(chalk.yellow('parametres locaux')).linefeed().indent()
      this.instructionArray(parameters.locals)
      this.str.deindent().linefeed()
    }
  }

  visitBinary(binary) {
    binary.left.accept(this)
    this.str.print(' ' + chalk.magenta(binary.op) + ' ')
    binary.right.accept(this)
  }

  visitRecordAccess(record) {
    record.record.accept(this)
    this.str.print(chalk.magenta('.') + chalk.white(record.property))
  }

  visitIdentifierRef(identifier) {
    this.str.print(chalk.white(identifier.ref))
  }

  visitArrayAccess(array) {
    array.array.accept(this)
    this.str.print(chalk.magenta('['))
    array.positions.forEach((position, i) => {
      if (i != 0)
        this.str.print(', ')
      position.accept(this)
    })
    this.str.print(chalk.magenta(']'))
  }

  visitUnary(unary) {
    this.str.print(chalk.magenta(unary.op))
    if (unary.op === 'non')
      this.str.print(' ')
    unary.expr.accept(this)
  }

  printValue(value) {
    switch(value.type) {
    case 'chaine':
    case 'string':
      return chalk.green('"' + value.value.replace(/"/g, '\\"').replace(/\n/g, '\\n') + '"')
    case 'caractere':
      return chalk.green("'" + value.value.replace(/"/g, '\\"').replace(/\n/g, '\\n') + "'")
    default:
      return chalk.blue(value.value)
    }
  }

  visitValue(value) {
    this.str.print(this.printValue(value))
  }

  visitSi(si) {
    this.str.print(chalk.yellow('si '))
    si.expr.accept(this)
    this.str.print(chalk.yellow(' alors')).linefeed().indent()
    this.instructionArray(si.vrai)
    if (si.faux) {
      this.str.deindent().print(chalk.yellow('sinon')).linefeed().indent()
      this.instructionArray(si.faux)
    }
    this.str.deindent().print(chalk.yellow('fin si'))
  }

  visitRoutineCall(routine) {
    this.str.print(chalk.white(routine.ref) + '(')
    routine.args.forEach((arg, i) => {
      if (i != 0)
        this.str.print(', ')
      arg.accept(this)
    })
    this.str.print(')')
  }

  visitPour(pour) {
    this.str.print(chalk.yellow('pour ') + chalk.white(pour.ref) + chalk.magenta(' ← '))
    pour.start.accept(this)
    this.str.print(chalk.yellow(' jusqu\'a '))
    pour.end.accept(this)
    if (pour.desc)
      this.str.print(chalk.yellow(' decroissant'))
    this.str.print(chalk.yellow(' faire')).linefeed().indent()
    this.instructionArray(pour.loop)
    this.str.deindent().print(chalk.yellow('fin pour'))
  }

  visitTantQue(tantque) {
    this.str.print(chalk.yellow(tantque.oneLoop ? 'faire' : 'tant que '))
    if (!tantque.oneLoop) {
      tantque.expr.accept(this)
      this.str.print(chalk.yellow(' faire'))
    }
    this.str.linefeed().indent()
    this.instructionArray(tantque.loop)
    this.str.deindent().print(chalk.yellow(tantque.oneLoop ? 'tant que ' : 'fin tant que'))
    if (tantque.oneLoop)
      tantque.expr.accept(this)
  }

  visitPointerAccess(pointer) {
    pointer.pointer.accept(this)
    this.str.print(chalk.magenta('↑'))
  }

  visitSelon(selon) {
    this.str.print(chalk.yellow('selon '))
    selon.test.accept(this)
    this.str.print(chalk.yellow(' faire')).linefeed().indent()

    this.str.startcol()
    this.instructionArray(selon.expressions, false)

    if (selon.either && selon.either.length) {
      this.str.newrow(chalk.yellow('autrement'))
      this.instructionArray(selon.either)
    }

    this.str.endcol();

    this.str.deindent().print(chalk.yellow('fin selon'))
  }

  visitSelonExpression(expr) {
    const header = expr.scalars.map(
      scalar => scalar.value instanceof ast.IdentifierRef ? scalar.value.ref : this.printValue(scalar.value)
    ).join(', ')
    this.str.newrow(header + chalk.yellow(':'))
    this.instructionArray(expr.expressions)
  }
}
