import {EmptyVisitor} from './empty.js'
import {JungleError} from '../jungle/error.js'
import * as ast from '../ast'
import * as types from '../types'
import * as routines from '../routines'

export class Scopes {
  constructor() {
    this.scopes = []
    this.pushScope()
    this.addType(new ast.Enum(['faux', 'vrai']).setName('booleen'))
    this.addType(new types.Caractere())
    this.addType(new types.Chaine())
    this.addType(new types.Entier())
    this.addType(new types.Reel())
    this.addType(new types.NoValue())
    this.addType(new types.NulPointer())
    this.addRoutine(new routines.Afficher(this.getType('chaine')))
    this.addRoutine(new routines.Allouer())
    this.addRoutine(new routines.Caractere(this.getType('caractere')))
    this.addRoutine(new routines.Chaine(this.getType('chaine')))
    this.addRoutine(new routines.Entier(this.getType('entier')))
    this.addRoutine(new routines.Liberer())
    this.addRoutine(new routines.Longueur(this.getType('entier'), this.getType('chaine')))
  }

  pushScope() {
    this.scopes.push({
      consts: {},
      types: {},
      values: {},
      variables: {},
      routines: {},
    })
  }

  popScope() {
    this.scopes.pop()
  }

  get scope() {
    return this.scopes[this.scopes.length - 1]
  }

  getConst(name) {
    for (let i = this.scopes.length - 1; i >= 0; --i) {
      if (name in this.scopes[i].consts)
        return this.scopes[i].consts[name]
    }
    throw new JungleError('Impossible de trouver la constante ' + name)
  }

  getType(name) {
    for (let i = this.scopes.length - 1; i >= 0; --i) {
      if (name in this.scopes[i].types)
        return this.scopes[i].types[name]
    }
    throw new JungleError('Impossible de trouver le type ' + name)
  }

  getVariable(name) {
    for (let i = this.scopes.length - 1; i >= 0; --i) {
      if (name in this.scopes[i].variables)
        return this.scopes[i].variables[name]
    }
    throw new JungleError('Impossible de trouver la variable ' + name)
  }

  getValue(value) {
    for (let i = this.scopes.length - 1; i >= 0; --i) {
      if (value in this.scopes[i].values)
        return this.scopes[i].values[value]
    }
    throw new JungleError('Impossible de trouver la valeur ' + value)
  }

  getRoutine(name) {
    for (let i = this.scopes.length - 1; i >= 0; --i) {
      if (name in this.scopes[i].routines)
        return this.scopes[i].routines[name]
    }
    throw new JungleError('Impossible de trouver la routine ' + name)
  }

  addConst(c) {
    if (!(c instanceof ast.Const))
      throw new JungleError('Le paramètre doit être une constante')

    if (c.name in this.scope.consts)
      throw new JungleError('Impossible de redéfinir la constante ' + c.name)

    this.scope.consts[c.name] = c
  }

  addType(type) {
    if (!(type instanceof ast.Type))
      throw new JungleError('Le paramètre doit être un type')

    if (type.name in this.scope.types)
      throw new JungleError('Impossible de redéfinir le type ' + type.name)

    this.scope.types[type.name] = type

    if (type instanceof ast.Enum)
      this.addValues(type)
  }

  addVariable(name, type) {
    if (name in this.scope.consts)
      throw new JungleError('Impossible de redéfinir la variable ' + name + ' (une constante a déjà le même nom)')
    if (name in this.scope.values)
      throw new JungleError('Impossible de redéfinir la variable ' + name + ' (l\'enum ' + this.scope.values[name].name + ' définit cette valeur)')
    if (name in this.scope.variables)
      throw new JungleError('Impossible de redéfinir la variable ' + name)

    this.scope.variables[name] = this.getType(type)
  }

  addValues(type) {
    type.values.forEach(value => this.addValue(type, value))
  }

  addValue(type, value) {
    if (value in this.scope.values)
      throw new JungleError('Impossible de redéfinir la valeur ' + value + ' dans le type ' + type.name + ' car déjà définie par le type ' +
        this.scope.values[value].name)

    this.scope.values[value] = type
  }

  addRoutine(routine) {
    if (routine.name in this.scope.routines)
      throw new JungleError('Impossible de redéfinir l\'algorithme ' + routine.name)

    this.scope.routines[routine.name] = routine
  }
}

export class Typer extends EmptyVisitor {
  visitAlgorithm(algo) {
    this.scopes = new Scopes()
    super.visitAlgorithm(algo)
    algo.typed = true
  }

  // Those need new scopes
  visitFunction(func) {
    this.scopes.addRoutine(func)
    this.scopes.pushScope()
    func.type = this.scopes.getType(func.type)
    this.scopes.addRoutine(new routines.Retourne(func))
    super.visitFunction(func)
    this.scopes.popScope()
  }
  visitProcedure(func) {
    this.scopes.addRoutine(func)
    this.scopes.pushScope()
    func.type = this.scopes.getType('<no value>')
    super.visitProcedure(func)
    this.scopes.popScope()
  }

  // Add values in scopes
  visitAlgoDecls(decls) {
    if (decls.consts.length)
      this.array(decls.consts)

    if (decls.types.length) {
      const prevMode = this.typeMode
      this.typeMode = 'register'
      this.array(decls.types)
      this.typeMode = 'check'
      this.array(decls.types)
      this.typeMode = prevMode
    }

    if (decls.variables.length)
      this.array(decls.variables)

    if (decls.routines && decls.routines.length)
      this.array(decls.routines)
  }

  visitConst(c) {
    c.type = this.scopes.getType(c.value.type)
    this.scopes.addConst(c)
  }
  visitArray(array) {
    if (this.typeMode === 'check') {
      array.type = this.scopes.getType(array.type)
      return
    }
    array.dimensions = array.dimensions.map(dimension => {
      if (!isNaN(+dimension))
        return dimension

      const c = this.scopes.getConst(dimension)

      if (c.value.type !== 'entier')
        throw new JungleError('Dans la définition du type tableau ' + array.name + ': ' + dimension + ' n\'est pas un entier')

      return c.value.value
    })
    this.scopes.addType(array)
  }
  visitEnum(e) {
    if (this.typeMode === 'check')
      return
    this.scopes.addType(e)
  }
  visitRecord(record) {
    if (this.typeMode === 'register') {
      this.scopes.addType(record)
      return
    }
    this.scopes.pushScope()
    super.visitRecord(record)
    this.scopes.popScope()
  }
  visitPointer(pointer) {
    if (this.typeMode === 'check') {
      pointer.pointOn = this.scopes.getType(pointer.pointOn)
      return
    }
    this.scopes.addType(pointer)
  }
  visitVariables(variables) {
    variables.names.forEach(name => this.scopes.addVariable(name, variables.type))
    variables.type = this.scopes.getType(variables.type)
  }

  // Check typing
  visitIdentifierRef(identifier) {
    try {
      const c = this.scopes.getConst(identifier.ref)
      identifier.resType = c.type
      identifier.value = c.value.value
    }
    catch(err) {
      try {
        identifier.resType = this.scopes.getValue(identifier.ref)
      }
      catch(err) {
        identifier.resType = this.scopes.getVariable(identifier.ref)
        identifier.writable = true
      }
    }
  }

  visitRoutineCall(routine) {
    routine.toCall = this.scopes.getRoutine(routine.ref)
    super.visitRoutineCall(routine)
    routine.resType = routine.toCall.type || this.scopes.getType('<no value>')

    routine.toCall.checkParameters(routine.args)
  }

  visitRecordAccess(record) {
    super.visitRecordAccess(record)
    if (!(record.record.resType instanceof ast.Record))
      throw new JungleError('Impossible d\'utiliser l\'opérateur . sur une expression qui n\'est pas un enregistrement (type ' + record.record.resType.name + ')')

    const property = record.record.resType.properties.find(variables => variables.names.indexOf(record.property) >= 0)
    if (!property)
      throw new JungleError('La propriété ' + record.property + ' n\'existe pas dans l\'enregistrement ' + record.record.resType.name)
    record.resType = property.type

    record.writable = record.record.writable
  }

  visitArrayAccess(array) {
    super.visitArrayAccess(array)

    const isArray = array.array.resType instanceof ast.Array
    const isChaine = array.array.resType instanceof types.Chaine

    if (!isArray && !isChaine)
      throw new JungleError('Impossible d\'utiliser l\'opérateur [] sur une expression qui n\'est pas un tableau ou une chaine (type ' + array.arrayType.name + ')')

    if (isArray && array.positions.length != array.array.resType.dimensions.length)
      throw new JungleError('Impossible d\'accéder à un tableau de type ' + array.array.resType.name + ' en utilisant ' +
        array.positions.length + ' dimensions alors qu\'il en comporte ' + array.array.resType.dimensions.length)
    else if (isChaine && array.positions.length != 1)
      throw new JungleError('Impossible d\'accéder à une chaine en utilisant ' + array.positions.length + ' dimensions')

    array.resType = isArray ? array.array.resType.type : this.scopes.getType('caractere')

    array.positions.forEach((pos, i) => {
      if (pos.resType !== this.scopes.getType('entier'))
        throw new JungleError('Le type de l\'expression pour la dimension ' + (i + 1) + ' lors de l\'accès à un tableau de type ' +
          array.resType.name + ' doit être un entier et non un ' + pos.resType.name)
    })

    array.writable = isArray && array.array.writable
  }

  visitValue(value) {
    value.resType = this.scopes.getType(value.type)
  }

  visitUnary(unary) {
    super.visitUnary(unary)
    switch (unary.op) {
    case 'non':
      if (unary.expr.resType.name !== 'booleen')
        throw new JungleError('L\'opérateur unaire non doit être utilisé sur un booleen')
      break
    case '+':
    case '-':
      if (unary.expr.resType !== this.scopes.getType('entier') && unary.expr.resType !== this.scopes.getType('reel'))
        throw new JungleError('L\'opérateur unaire ' + unary.op + ' doit être utilisé sur des valeur numériques (entier/reel) mais ' +
          'a été appliqué sur un ' + unary.expr.resType.name)
      break
    default:
      throw new JungleError('Opérateur unaire inconnu : ' + unary.op)
    }

    unary.resType = unary.expr.resType
  }

  visitBinary(binary) {
    super.visitBinary(binary)

    const opPossibilities = {
      arith: {
        entry: ['entier', 'reel', ['entier', 'reel']],
        out: 'last',
      },
      entier: 'entier',
      plus: {
        entry: ['entier', 'reel', 'chaine', ['caractere', 'chaine'], ['entier', 'reel']],
        out: 'last',
      },
      cmp: {
        entry: ['entier', 'reel', 'chaine', ['caractere', 'chaine'], ['entier', 'reel']],
        out: 'booleen',
      },
      bool: 'booleen',
    }

    function resolvePossibilites(scopes, possibilities, left, right) {
      if (typeof possibilities !== 'object')
        possibilities = {entry: possibilities, out: 'last'}
      if (!Array.isArray(possibilities.entry))
        possibilities.entry = [possibilities.entry]

      const entry = possibilities.entry.find(entry => {
        if (typeof entry === 'string')
          entry = [entry]

        entry = entry.map(type => scopes.getType(type))

        return entry.indexOf(left) >= 0 && entry.indexOf(right) >= 0
      })

      if (!entry) {
        throw new JungleError('Mauvais opérandes lors de l\'utilisation de l\'opérateur binaire ' + binary.op + '. Un type parmis les suivants est attendu : ' +
          possibilities.entry.filter(p => typeof p === 'string').join(', ') + '. Les types utilisés sont : ' + left.name + ' et ' + right.name)
      }

      if (possibilities.out === 'last')
        return Array.isArray(entry) ? scopes.getType(entry[1]) : scopes.getType(entry)
      return scopes.getType(possibilities.out)
    }

    let choice = null

    switch(binary.op) {
    case 'div':
    case 'mod':
      choice = 'entier'
      break

    case '+':
      choice = 'plus'
      break

    case '-':
    case '*':
    case '/':
      choice = 'arith'
      break

    case '=':
    case '<>':
      const isEnum = binary.left.resType instanceof ast.Enum
      const isPointer = binary.left.resType instanceof ast.Pointer
      const sameTypes = binary.left.resType === binary.right.resType
      if ((isEnum && sameTypes) ||
        (isPointer && (sameTypes ||
          binary.left.resType instanceof types.NulPointer ||
          binary.right.resType instanceof types.NulPointer))) {
        binary.resType = this.scopes.getType('booleen')
        return
      }
    case '<':
    case '>':
    case '<=':
    case '>=':
      choice = 'cmp'
      break

    case 'et':
    case 'ou':
    case 'oue':
      choice = 'bool'
      break

    case '←':
      if (!binary.left.writable)
        throw new JungleError('Impossible d\'assigner une valeur dans une expression non assignable de type ' + binary.left.resType.name)
      if (binary.left.resType !== binary.right.resType && binary.right.resType.name !== '!null' && !(binary.left.resType instanceof ast.Pointer))
        throw new JungleError('Impossible d\'assigner une valeur de type ' + binary.right.resType.name + ' dans une expression de type ' + binary.left.resType.name)
      binary.resType = this.scopes.getType('<no value>')
      return

    default:
      throw binary.location, new JungleError('Opérateur ' + binary.operator + ' inconnu')
    }

    binary.resType = resolvePossibilites(this.scopes, opPossibilities[choice], binary.left.resType, binary.right.resType)
  }

  visitSi(si) {
    super.visitSi(si)

    if (si.expr.resType !== this.scopes.getType('booleen'))
      throw new JungleError('si prend une expression booléenne et non un ' + si.expr.resType.name)

    si.resType = this.scopes.getType('<no value>')
  }

  visitPour(pour) {
    super.visitPour(pour)

    pour.refType = this.scopes.getVariable(pour.ref)

    const entier = this.scopes.getType('entier')

    if (pour.refType !== entier)
      throw new JungleError('pour prend une variable de type entier et non un ' + pour.refType.name)

    if (pour.start.resType !== entier)
      throw new JungleError('pour prend une expression entière comme valeur de départ et non un ' + pour.start.resType.name)

    if (pour.end.resType !== entier)
      throw new JungleError('pour prend une expression entière comme valeur de fin et non un ' + pour.end.resType.name)

    pour.resType = this.scopes.getType('<no value>')
  }

  visitTantQue(tantque) {
    super.visitTantQue(tantque)

    if (tantque.expr.resType !== this.scopes.getType('booleen'))
      throw new JungleError('tant que prend une expression booléenne et non un ' + tantque.expr.resType.name)

    tantque.resType = this.scopes.getType('<no value>')
  }

  visitPointerAccess(pointer) {
    super.visitPointerAccess(pointer)

    if (!(pointer.pointer.resType instanceof ast.Pointer))
      throw new JungleError('Impossible de déréférencer un type qui n\'est pas un pointeur')

    pointer.resType = pointer.pointer.resType.pointOn
    pointer.writable = true
  }

  visitSelon(selon) {
    super.visitSelon(selon)

    if (selon.test.resType !== this.scopes.getType('entier') &&
      selon.test.resType !== this.scopes.getType('caractere') &&
      !(selon.test.resType instanceof ast.Enum))
      throw new JungleError('selon attend une expression de type entier, caractere ou enum et non un ' + selon.test.resType)

    selon.expressions.forEach(expr => {
      if (expr.resType !== selon.test.resType)
        throw new JungleError('selon a une expression de type ' + selon.test.resType.name +
          ' mais les scalaires de cette liste sont du type ' + expr.resType.name, expr.scalars[0].location)
    })
  }

  visitSelonExpression(expr) {
    super.visitSelonExpression(expr)

    expr.resType = expr.scalars.reduce((acc, scalar) => {
      if (acc.resType !== scalar.resType)
        throw new JungleError('tous les scalaires dans la liste du selon doivent être du même type, ici ' +
          acc.resType.name + ' != ' + scalar.resType.name, scalar.location)

      return scalar
    }).resType
  }

  visitScalar(scalar) {
    super.visitScalar(scalar)

    if (scalar.value.resType !== this.scopes.getType('entier') &&
      scalar.value.resType !== this.scopes.getType('caractere') &&
      !(scalar.value.resType instanceof ast.Enum))
      throw new JungleError('les scalaires dans selon doivent être de type entier, caractere ou enum et non un ' + scalar.value.resType)

    if (scalar.value.writable)
      throw new JungleError('impossible d\'utiliser une variable comme scalaire dans selon')

    scalar.resType = scalar.value.resType
  }
}
